
#include "GlutWindow.h"
#include "ArcBall.h"

#include <math.h>
#include <GL/glut.h>
#include <stdio.h>

#define DEG2RAD(a) ((a)/57.295)

#define ZOOM_MIN (0.05)
#define ZOOM_MAX (10.0)

#define VECTOR_FILENAME "circular.raw"

CGlutWindow::CGlutWindow(void)
{
	m_pArcBall = new CArcBall();

	glutInitWindowSize (800,600); 
	glutInitWindowPosition (100, 100);
	glutCreateWindow ("GLUT Window");

	m_dFieldOfView = 30.0;
	m_nWidth = 1;
	m_nHeight = 1;
	
	m_dZoom = 1.0;
	m_dTranslateX = 0.0;
	m_dTranslateY = 0.0;
	m_nMouseState = 0;

	m_nSkipX = 2;
	m_nSkipY = 2;
	m_nSkipZ = 2;
	m_nTime = 0;
	m_fTimeStep = 0.1;
	
	m_nNumTraces = 120;
	m_nNumPositionsPerTrace = 200;

	m_pVectorfield = NULL;
	m_pParticleTraces = NULL;

	loadVectorField();

	m_dRadius = 0.5*sqrt(float(m_nXdim*m_nXdim + m_nYdim*m_nYdim + m_nZdim*m_nZdim));
	m_vecCameraPosition.set(0.0, 0.0, 0.0,1.0);

	m_dCenter = m_dRadius/ sin(DEG2RAD(m_dFieldOfView)/2.0);

	m_bDrawParticleTraces = true;

	m_bDrawArrowPlots = false;


	initializeGL();
}

void CGlutWindow::loadVectorField() {
	
	FILE *pFile = fopen(VECTOR_FILENAME,"rb");

	if (pFile == NULL) exit(-2);
	
	fread(&m_nXdim,sizeof(unsigned int),1,pFile);
	fread(&m_nYdim,sizeof(unsigned int),1,pFile);
	fread(&m_nZdim,sizeof(unsigned int),1,pFile);
	fread(&m_nNumTimeSteps,sizeof(unsigned int),1,pFile);

	int nNumBytesToRead = m_nXdim * m_nYdim * m_nZdim * m_nNumTimeSteps * 3;
	m_pVectorfield = new float[nNumBytesToRead];

	int nNumBytesRead = fread(m_pVectorfield,sizeof(float),nNumBytesToRead,pFile);

	if (nNumBytesRead != nNumBytesToRead) {
		exit(-1);
	}

	calculateParticleTraces();
	
}

void CGlutWindow::calculateParticleTraces() {

	int nNumFloats = m_nNumPositionsPerTrace * m_nNumTraces * 3;
	
	if (NULL != m_pParticleTraces) delete [] m_pParticleTraces;

	m_pParticleTraces = new float[nNumFloats];

	int idx = 0;
	CVector vec;
	for(int i = 0; i < m_nNumTraces; ++i) {
		vec[0] = float(rand()%m_nXdim);
		vec[1] = float(rand()%m_nYdim);
		vec[2] = float(rand()%m_nZdim);

		for(int j = 0; j < m_nNumPositionsPerTrace; ++j) {
			m_pParticleTraces[idx++] = vec[0];
			m_pParticleTraces[idx++] = vec[1];
			m_pParticleTraces[idx++] = vec[2];
			nextPosition(vec);
		}
	}
}

void CGlutWindow::nextPosition(CVector &vecP) {

	int nX = int(vecP[0]);
	int nY = int(vecP[1]);
	int nZ = int(vecP[2]);
	
	if ((nX < 0) || (nX > m_nXdim-2)) return;
	if ((nY < 0) || (nY > m_nYdim-2)) return;
	if ((nZ < 0) || (nZ > m_nZdim-2)) return;

	float fX = vecP[0] - float(nX);
	float fY = vecP[1] - float(nY);
	float fZ = vecP[2] - float(nZ);

	int idx000 = ((( nZ    * m_nYdim)+nY  ) * m_nXdim + nX  )* 3;
	int idx001 = ((( nZ    * m_nYdim)+nY  ) * m_nXdim + nX+1)* 3;
	int idx010 = ((( nZ    * m_nYdim)+nY+1) * m_nXdim + nX  )* 3;
	int idx011 = ((( nZ    * m_nYdim)+nY+1) * m_nXdim + nX+1)* 3;
	int idx100 = ((((nZ+1) * m_nYdim)+nY  ) * m_nXdim + nX  )* 3;
	int idx101 = ((((nZ+1) * m_nYdim)+nY  ) * m_nXdim + nX+1)* 3;
	int idx110 = ((((nZ+1) * m_nYdim)+nY+1) * m_nXdim + nX  )* 3;
	int idx111 = ((((nZ+1) * m_nYdim)+nY+1) * m_nXdim + nX+1)* 3;
	
	CVector v000(m_pVectorfield[idx000],m_pVectorfield[idx000+1],m_pVectorfield[idx000+2]);
	CVector v001(m_pVectorfield[idx001],m_pVectorfield[idx001+1],m_pVectorfield[idx001+2]);
	CVector v010(m_pVectorfield[idx010],m_pVectorfield[idx010+1],m_pVectorfield[idx010+2]);
	CVector v011(m_pVectorfield[idx011],m_pVectorfield[idx011+1],m_pVectorfield[idx011+2]);
	CVector v100(m_pVectorfield[idx100],m_pVectorfield[idx100+1],m_pVectorfield[idx100+2]);
	CVector v101(m_pVectorfield[idx101],m_pVectorfield[idx101+1],m_pVectorfield[idx101+2]);
	CVector v110(m_pVectorfield[idx110],m_pVectorfield[idx110+1],m_pVectorfield[idx110+2]);
	CVector v111(m_pVectorfield[idx111],m_pVectorfield[idx111+1],m_pVectorfield[idx111+2]);

	CVector vx00 = (1.0-fX) * v000 + fX * v100;
	CVector vx01 = (1.0-fX) * v001 + fX * v101;
	CVector vx10 = (1.0-fX) * v010 + fX * v110;
	CVector vx11 = (1.0-fX) * v011 + fX * v111;

	CVector vxy0 = (1.0-fY) * vx00 + fY * vx10;
	CVector vxy1 = (1.0-fY) * vx01 + fY * vx11;

	CVector vxyz = (1.0-fZ) * vxy0 + fZ * vxy1;

	vecP += m_fTimeStep * vxyz;

}

CGlutWindow::~CGlutWindow(void)
{
	if (NULL != m_pArcBall) delete m_pArcBall;
}


void CGlutWindow::renderFrame() {

	double dAspectRatio = double(m_nWidth)/double(m_nHeight);

	// The usual OpenGL stuff to clear the screen and set up viewing.
	//glClearColor(0.5,0.5,1.0,0.0);
	glClearColor(1.0,1.0,1.0,0.0);
	//glClearColor(0.0,0.0,0.0,0.0);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	GLfloat fTop, fRight, fNear, fFar;

	fNear   = float(m_dCenter - m_dRadius);
	fFar    = float(m_dCenter + m_dRadius);

	if (dAspectRatio > 1.0) {
		fRight = fNear * float(tan(DEG2RAD(m_dFieldOfView)/2.0) /m_dZoom);
		fTop   = fRight * float(dAspectRatio);		
	} else {
		fTop   = fNear * float(tan(DEG2RAD(m_dFieldOfView)/2.0) /m_dZoom);
		fRight = fTop / float(dAspectRatio);
	} 

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum( -fTop, fTop, -fRight, fRight, fNear, 10*fFar);

	CMatrix matModel = m_pArcBall->getRotatonMatrix().getInverse();

	double matrix[16];
	matModel.get(matrix[ 0], matrix[ 4], matrix[ 8], matrix[12],
		matrix[ 1], matrix[ 5], matrix[ 9], matrix[13],
		matrix[ 2], matrix[ 6], matrix[10], matrix[14],
		matrix[ 3], matrix[ 7], matrix[11], matrix[15]);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslated(m_dTranslateX * m_dRadius/m_dZoom, m_dTranslateY*m_dRadius/m_dZoom, -m_dCenter);
	glMultMatrixd(matrix);
	glTranslated(m_vecCameraPosition[0], m_vecCameraPosition[1], m_vecCameraPosition[2]);

	glClear(GL_COLOR_BUFFER_BIT);

    renderGeometry();

	glutSwapBuffers();
}


void CGlutWindow::resize(int width, int height) {
	m_nWidth = width; 
	m_nHeight = height;
	glViewport(0, 0, (GLsizei) width, (GLsizei) height);

	m_pArcBall->place(CPosition(width/2, height/2), 
		0.5 * sqrt(double(width * width + height * height)));
}

void CGlutWindow::keyEvent(unsigned char key,int x,int y){

	switch(key) {
		case 'x':
			m_nSkipX++;
			break;
		case 'X':
			m_nSkipX--;
			if (m_nSkipX < 0) m_nSkipX = 0;
			break;
			
		case 'y':
			m_nSkipY++;
			break;
		case 'Y':
			m_nSkipY--;
			if (m_nSkipY < 0) m_nSkipY = 0;
			break;
			
		case 'z':
			m_nSkipZ++;
			break;
		case 'Z':
			m_nSkipZ--;
			if (m_nSkipZ < 0) m_nSkipZ = 0;
			break;
		case 't':
			m_nTime = (m_nTime-1)%m_nNumTimeSteps;
			break;

		case 'T':
			m_nTime = (m_nTime+1)%m_nNumTimeSteps;
			break;

		case 'a':
			m_bDrawArrowPlots = !m_bDrawArrowPlots;
			break;

		case 'p':
			m_bDrawParticleTraces = !m_bDrawParticleTraces;
			break;

		case 'c':
			calculateParticleTraces();
			break;

	}
	
		glutPostRedisplay();
}

void CGlutWindow::mouseButtonEvent(int button, int state,int x,int y){

	switch(state) {
		case GLUT_DOWN:
			{
				m_vecMouseDown.set(double(x)/double(m_nWidth),double(y)/double(m_nHeight));

				switch (button) {
					case GLUT_LEFT_BUTTON:
						{
							m_nMouseState |= (MBUTTON_LEFT);
							m_pArcBall->setMouse( CPosition(x, m_nHeight-1-y));
							m_pArcBall->startDragging();
						}
						break;
					case GLUT_RIGHT_BUTTON:
						{
							m_nMouseState |= (MBUTTON_RIGHT);	
							m_bDragging = false;
							m_bZooming = true;
						}
						break;
					case GLUT_MIDDLE_BUTTON:
						{
							m_nMouseState |= (MBUTTON_MIDDLE);						
							m_bDragging = true;
							m_bZooming = false;
						}
						break;
					default:
						break;
				}
				break;
		case GLUT_UP:
			{
				switch (button) {
					case GLUT_LEFT_BUTTON:
						m_nMouseState ^= MBUTTON_LEFT;
						m_pArcBall->finishDragging();
						break;
					case GLUT_RIGHT_BUTTON:
						m_nMouseState ^= MBUTTON_RIGHT;
						break;
					case GLUT_MIDDLE_BUTTON:
						m_nMouseState ^= MBUTTON_MIDDLE;
						break;
					default:
						break;
				}
				m_bDragging = false;
				m_bZooming  = false;
				}
			}


	}
}

void CGlutWindow::mouseMoveEvent(int x,int y){

	 
	CPosition pos(double(x)/double(m_nWidth),double(y)/double(m_nHeight));

	CPosition drag = pos-m_vecMouseDown;

	if (m_nMouseState & (MBUTTON_LEFT)) {
		m_pArcBall->setMouse(CPosition(x, m_nHeight-1- y));
		glutPostRedisplay();

	} else if ((m_bDragging) && (m_nMouseState & MBUTTON_MIDDLE)) {
		
		m_dTranslateX +=  2.0 * drag[0];
		m_dTranslateY += -2.0 * drag[1];

		if (m_dTranslateX >  m_dZoom) m_dTranslateX =  m_dZoom;
		if (m_dTranslateX < -m_dZoom) m_dTranslateX = -m_dZoom;
		if (m_dTranslateY >  m_dZoom) m_dTranslateY =  m_dZoom;
		if (m_dTranslateY < -m_dZoom) m_dTranslateY = -m_dZoom;
		glutPostRedisplay();
	} else if ((m_bZooming) && (m_nMouseState & MBUTTON_RIGHT )) {
			
		double dZoom = 10.0*drag[0]; 
		m_dZoom += dZoom;
		if (m_dZoom < ZOOM_MIN) m_dZoom = ZOOM_MIN;
		if (m_dZoom > ZOOM_MAX) m_dZoom = ZOOM_MAX;
		if (m_dTranslateX >  m_dZoom) m_dTranslateX =  m_dZoom;
		if (m_dTranslateX < -m_dZoom) m_dTranslateX = -m_dZoom;
		if (m_dTranslateY >  m_dZoom) m_dTranslateY =  m_dZoom;
		if (m_dTranslateY < -m_dZoom) m_dTranslateY = -m_dZoom;
		glutPostRedisplay();
	} 
	m_vecMouseDown = pos;
}

void CGlutWindow::idle(){
}

void CGlutWindow::initializeGL()
{
    GLfloat ambient[] =
	{0.4f, 0.4f, 0.4f, 1.0f};
    GLfloat diffuse[] =
	{1.0f, 1.0f, 1.0f, 1.0f};
    GLfloat specular[] =
	{0.5f, 0.5f, 0.5f, 1.0f};
    GLfloat position[] =
	{1.0f, 1.0f, 1.0f, 1.0f};
    
    GLfloat lmodel_ambient[] =
	{0.2f, 0.2f, 0.2f, 1.0f};
    GLfloat local_view[] =
	{0.0f};
    
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, specular);
    glLightfv(GL_LIGHT0, GL_POSITION, position);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
    glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, local_view);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_DEPTH_TEST);
    
    /* Select Color and Projection*/
    glClearColor(0.0, 0.0, 0.0, 0.0);
    //glColor3f(1.0, 1.0, 1.0);
    glShadeModel(GL_SMOOTH);
	
	
}


void CGlutWindow::renderGeometry() {

	glDisable(GL_LIGHTING);
	glLineWidth(1.0);

	float fHalfX = float(m_nXdim)/2.0f;
	float fHalfY = float(m_nYdim)/2.0f;
	float fHalfZ = float(m_nZdim)/2.0f;

	// RENDER THE BOUNDING BOX...
	glColor4d(0.0,0.0,0.0,1.0);
	glBegin(GL_LINE_LOOP);
	{
		glVertex3f(-fHalfX,-fHalfY,-fHalfZ);
		glVertex3f( fHalfX,-fHalfY,-fHalfZ);
		glVertex3f( fHalfX, fHalfY,-fHalfZ);
		glVertex3f(-fHalfX, fHalfY,-fHalfZ);
	}
	glEnd();

	glBegin(GL_LINE_LOOP);
	{
		glVertex3f(-fHalfX,-fHalfY,fHalfZ);
		glVertex3f( fHalfX,-fHalfY,fHalfZ);
		glVertex3f( fHalfX, fHalfY,fHalfZ);
		glVertex3f(-fHalfX, fHalfY,fHalfZ);
	}
	glEnd();

	glBegin(GL_LINES);
	{
		glVertex3f(-fHalfX,-fHalfY,-fHalfZ);
		glVertex3f(-fHalfX,-fHalfY,fHalfZ);

		glVertex3f( fHalfX,-fHalfY,-fHalfZ);
		glVertex3f( fHalfX,-fHalfY,fHalfZ);

		glVertex3f( fHalfX, fHalfY,-fHalfZ);
		glVertex3f( fHalfX, fHalfY,fHalfZ);

		glVertex3f(-fHalfX, fHalfY,-fHalfZ);
		glVertex3f(-fHalfX, fHalfY,fHalfZ);
	}
	glEnd();

	if (m_bDrawArrowPlots) {
	// RENDER THE ARROW PLOT

	float fScale = 1.0;

	glBegin(GL_LINES);
	glColor3d(1.0,0.0,0.0);

	for(int z = 0; z < m_nZdim; z+= m_nSkipZ) {
			for(int y = 0; y < m_nYdim; y+= m_nSkipY) {
				for(int x = 0; x < m_nXdim; x+= m_nSkipX) {

					int idx = (((m_nTime * m_nZdim + z) * m_nYdim + y) * m_nXdim + x) * 3;
					float vx = m_pVectorfield[idx];
					float vy = m_pVectorfield[idx+1];
					float vz = m_pVectorfield[idx+2];

					float fx = float(x) - fHalfX;
					float fy = float(y) - fHalfY;
					float fz = float(z) - fHalfZ;

					glVertex3f(fx,fy,fz);
					glVertex3f(fx+vx,fy+vy,fz+vz);
					

				}
			}
		}

	glEnd();

	}

	if (m_bDrawParticleTraces) {
		int idx = 0;
		for(int i = 0; i < m_nNumTraces; ++i) {
			glBegin(GL_LINE_STRIP);
			glColor3f(float(rand()%255)/255.0f,float(rand()%255)/255.0f,float(rand()%255)/255.0f);
			for(int j = 0; j < m_nNumPositionsPerTrace; ++j) {
				glVertex3f(m_pParticleTraces[idx]-fHalfX,m_pParticleTraces[idx+1]-fHalfY,m_pParticleTraces[idx+2]-fHalfZ);
				idx += 3;
			}
			glEnd();

		}
	}


}