// TerraGen.cpp : Defines the entry point for the console application.
//

#include "GlutWindow.h"
#include <GL/glut.h>
#include <time.h>
#include <stdio.h>

CGlutWindow *g_pMainWindow = 0;

void display(void)                              { g_pMainWindow->renderFrame();}
void idle(void)                                 { g_pMainWindow->idle();}
void reshape(int width, int height)             { g_pMainWindow->resize(width, height);}
void keyboard(unsigned char key, int x, int y)  { g_pMainWindow->keyEvent(key,x,y);}
void mouse(int button, int state, int x, int y) { g_pMainWindow->mouseButtonEvent(button, state,x,y);}
void move(int x, int y)                         { g_pMainWindow->mouseMoveEvent(x,y);}

void createVectorField() {

	unsigned int nXdim = 50; 
	unsigned int nYdim = 50; 
	unsigned int nZdim = 32; 
	unsigned int nNumTimeSteps = 10;

	float *pBuffer = new float[nXdim*nYdim*nZdim*nNumTimeSteps*3];

	float fCenterX = float(nXdim)/2.0f;
	float fCenterY = float(nYdim)/2.0f;
	float fCenterZ = float(nZdim)/2.0f;

	int idx = 0;
	for(int t = 0; t < nNumTimeSteps; t++) {
		for(int z = 0; z < nZdim; z++) {
			for(int y = 0; y < nYdim; y++) {
				for(int x = 0; x < nXdim; x++) {

					pBuffer[idx++] = 0.1 * -( float(y)-fCenterY);
					pBuffer[idx++] = 0.1 *   (float(x)-fCenterX);
					pBuffer[idx++] = 0.1 *    float(t);
				}
			}
		}
	}

	FILE *pFile = fopen("circular.raw","wb");
	fwrite(&nXdim,sizeof(unsigned int),1,pFile);
	fwrite(&nYdim,sizeof(unsigned int),1,pFile);
	fwrite(&nZdim,sizeof(unsigned int),1,pFile);
	fwrite(&nNumTimeSteps,sizeof(unsigned int),1,pFile);
	fwrite(pBuffer,sizeof(float),nXdim*nYdim*nZdim*nNumTimeSteps*3,pFile);
	fclose(pFile);
}


int main(int argc, char* argv[])
{
	printf("Command keys:\n\n");
	printf("a - display/hide arrow plot\n\n");
	printf("  x/X increase/decrease arrow density in x-direction\n");
	printf("  y/Y increase/decrease arrow density in y-direction\n");
	printf("  z/Z increase/decrease arrow density in z-direction\n\n\n");
	printf("p - display/hide particle traces\n\n");
	printf("  c - recalculate random particle traces\n\n\n");
	printf("FIXME: Particle traces are calculated in the\n");
	printf("       first timestep only, not in the dynamic vector field!\n");

	//createVectorField();

	glutInit(&argc, argv);
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB);

	g_pMainWindow = new CGlutWindow();

	glutDisplayFunc(display); 
    glutReshapeFunc(reshape);
    glutKeyboardFunc(keyboard);
    glutMouseFunc(mouse);
    glutMotionFunc(move);
	glutIdleFunc(idle);
	glutMainLoop();

	return 0;
}
